# cython: language_level=2
# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""_runtime_pymisc.pxd provides miscellaneous Python-related utilities for _runtime_*.pyx"""

# PyExc and pyexc_fetch/pyexc_restore provide non-noisy way to save/restore
# current python exception.
cdef extern from "Python.h":
    """
    typedef struct PyExc {
        PyObject *pyexc_type;
        PyObject *pyexc_value;
        PyObject *pyexc_traceback;
    } PyExc;

    static inline void pyexc_fetch(PyExc *e) {
        PyErr_Fetch(&e->pyexc_type, &e->pyexc_value, &e->pyexc_traceback);
    }

    static inline void pyexc_restore(PyExc e) {
        PyErr_Restore(e.pyexc_type, e.pyexc_value, e.pyexc_traceback);
    }
    """
    struct PyExc:
        pass
    void pyexc_fetch(PyExc*)
    void pyexc_restore(PyExc)
